%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Version:               1.0                                                                                   
%Date:                  27 July 2020                       
%Authors:               Timothy Chadza, Konstantinos Kyriakopoulos & Sangarapillai Lambotharan                 
%Affiliation:           Loughborough University,Loughborough,LE11 3TU, UK                                    
%Corresponding Author:  Timothy Chadza, tchadza@poly.ac.mw
%Copyright:             This is published under GNU GENERAL PUBLIC LICENSE Version 3. If you use this code,   
%                       please, cite our Elsevier paper: 
%                       Timothy Chadza, Konstantinos G. Kyriakopoulos and Sangarapillai Lambotharan,
%                       Analysis of hidden Markov model learning algorithms for the detection and prediction 
%                       of multi-stage network attacks, Future Generation Computer Systems, Volume 108,2020,
%                       Pages 636-64, https://doi.org/10.1016/j.future.2020.03.014 
%Credits:               One of the algorithms used, spectral algorithm, was developed by Carl Mattfield. In 
%                       our work, we have used this spectral algorithm in its entirety and only added two
%                       lines at the end to obtain the HMM transitions matrices. For citation use of the 
%                       spectral algorithm work, please use, C. Mattfield, "Implementing Spectral Methods for
%                       Hidden Markov Models with Real-Valued Emissions",arXiv:1404.7472, 2014. 
%Brief Description:     Part 1: Predefined variables
%                               These can be modified accordingly.
%                       Part 2: Reading Files
%                               Other Snort alerts can be used, but preporocessing has to be modified         
%                               accordingly.
%                       Part 3: Preprocessing
%                               Fields of Snort raw files are extracted from raw files. The processes include: 
%                               removing duplication and extracting essential fields.
%                       Part 4: Training
%                               For training, the maximum likelihood estiamte is used in a supervised learning 
%                               fashion, furthermore, unsupervised training algorithms, namely, Baum-Welch, 
%                               segmental K-means (Viterbi training), differential evolution, spectral,
%                               K-means with and without phases are used. Addionally, hybrid techniques in 
%                               conjunction with either Baum-Welch or seqgmental K-means have been used.
%                               The entire dataset for the DARPA 2000 inside network has been used for 
%                               training 
%                       Part 5: Evaluation using A, B and alpha (method 1)
%                               The DMZ dataset has been used for evaluation. In the evaluation process, two 
%                               methods have been used. For this method, the A and B parameters, together with 
%                               the alpha(forward variable) have been used to detemine the next probable state 
%                               and next observation simultaneously. A window size has been selected and then 
%                               increased in steps till the maximum suggested size. Each time, the accuracy is 
%                               computed for all states (AS), currrent state (CS), next state (NS) and next 
%                               observation (NO). Two prediction levels have been used for NS and NO.
%                       Part 6: Evaluation using A and alpha for NS, and B and alpha for NO (method 2)
%                               in method 1, where the possible number of estimates is a function of N*M, for 
%                               method 2, the possible number of estimates for NS and NO is N and M, 
%                               respectively. As in method 1, a window size has been selected and then 
%                               increased in steps upto the maximum suggested size. Each time, the accuracy is 
%                               computed for AS, CS, NS and NO, where two prediction levels have been used for 
%                               NS and NO.
%                       Part 7: Plotting results
%                               Results have been presented as line graphs with the accuracy of all 
%                               techniques being plotted against a varying alert sample window size.
%                               The optimum window size can be determine from the results and critically 
%                               examined using a fixed window size.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clear;  clc;% clear variables and then clear screen

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Part 1:  Reading Files                                                                  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Read files containing alerts. Alert Name indicates the total number of alerts
Raw_Data_Inside1= fileread('alert_1inside2306');
Raw_Data_DMZ1 = fileread('alert_1dmz2949'); 


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Part 2: Predefined Variables
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Maxiter=500; %Maximum number of iterations
Tol=1e-6; %Convergence threshold
Eps=1e-10;%Small number to assign to zeros in matrices
Start=50; %Sample alert window size of first experiment
Step=50; %Step size to use fro varying sample alert window size
Stop=900; %Sample alert window size of last experiment

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Part 3: Preprocessing                                                                      
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fprintf('Step 1 of 4:Preprocessing\n');
[Inside1_Nodup_Time,DMZ1_Nodup_Time,L1,L2,Tr_Data_Inside1,Ev_Data_DMZ1,Inside1_Symbols,DMZ1_Symbols,Phase_Points1_Inside1,Phase_Points1_DMZ1]=preprocess(Raw_Data_Inside1,Raw_Data_DMZ1);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Part 4: Training                                                                                 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fprintf('Step 2 of 4:Training\n');
%Define the training and Evaluation Data
Training_Data=Tr_Data_Inside1;
Evaluation_Data=Ev_Data_DMZ1;

M1=max(Training_Data(:,1));%Number of observations in Training Data
M2=max(Evaluation_Data(:,1));%Number of observations in Evaluation Data
N=max(Training_Data(:,2)); %Number of states
%create sequence and states row matrices
Training_Seq(1,:)=Training_Data(1:end,1);
Training_States(1,:)=Training_Data(1:end,2);
Evaluation_Seq(1,:)=Evaluation_Data(1:end,1);
Evaluation_States(1,:)=Evaluation_Data(1:end,2);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Method 1: Using SL from MATLAB (hmmestimate)
%calculates the MLE estimate of A and B from sequence with known states
[estA_SL, B_SL] = hmmestimate(Training_Seq,Training_States);
A_SL=normalise_rows(estA_SL,Eps); %Check if there is state transition of zero and scale
%%%%%%%%%%%%%%%%%%%%BW
[estA1_SL_BW, B_SL_BW] = hmmtrain(Training_Seq,A_SL,B_SL,'ALGORITHM','BaumWelch','Maxiterations',Maxiter,'Tolerance',Tol);
A_SL_BW=normalise_rows(estA1_SL_BW,Eps);
%%%%%%%%%%%%%%%%%%%%SKM
[estA1_SL_SKM, B_SL_SKM] = hmmtrain(Training_Seq,A_SL,B_SL,'ALGORITHM','Viterbi','Maxiterations',Maxiter,'Tolerance',Tol,'Pseudotransitions',A_SL);
A_SL_SKM=normalise_rows(estA1_SL_SKM,Eps);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Method 2: %Use three standard techniques for BW
%2A: Uniform initialisation 
A_Unif=1/N*ones(N,N);
B_Unif=1/M1*ones(N,M1);
%%%%%%%%%%%%%%%%%%%%BW
[estA_Unif_BW, B_Unif_BW] = hmmtrain(Training_Seq,A_Unif,B_Unif,'ALGORITHM','BaumWelch','Maxiterations',Maxiter,'Tolerance',Tol);
A_Unif_BW=normalise_rows(estA_Unif_BW,Eps);%Check if there is state transition of zero and scale
%%%%%%%%%%%%%%%%%%%%SKM
[estA1_Unif_SKM, B_Unif_SKM] = hmmtrain(Training_Seq,A_Unif,B_Unif,'ALGORITHM','Viterbi','Maxiterations',Maxiter,'Tolerance',Tol,'Pseudotransitions',A_Unif);
A_Unif_SKM=normalise_rows(estA1_Unif_SKM,Eps);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%2B: Random initialisation
load('ABPi'); %Load predefined random A, B and Pi. Uncomment line below to generate new ones if required
% [A_Rand,B_Rand,Pi_Rand]=initialise_random(M1,N);%obtain initial A, B and Pi
%%%%%%%%%%%%%%%%%%%%BW
[EstA_Rand_BW, B_Rand_BW] = hmmtrain(Training_Seq,A_Rand,B_Rand,'ALGORITHM','BaumWelch','Maxiterations',Maxiter,'Tolerance',Tol);
A_Rand_BW=normalise_rows(EstA_Rand_BW,Eps);%Check if there is state tA_SA_HMMransition of zero and scale
%%%%%%%%%%%%%%%%%%%%SKM
[estA1_Rand_SKM, B_Rand_SKM] = hmmtrain(Training_Seq,A_Rand,B_Rand,'ALGORITHM','Viterbi','Maxiterations',Maxiter,'Tolerance',Tol,'Pseudotransitions',A_Rand);
A_Rand_SKM=normalise_rows(estA1_Rand_SKM,Eps);

%2C: Count based initialisation
[A_Count,B_Count]=count_based(M1,N,Training_Data);%obtain initial A and B
%%%%%%%%%%%%%%%%%%%%BW
[estA_Count_BW, B_Count_BW] = hmmtrain(Training_Seq,A_Count,B_Count,'ALGORITHM','BaumWelch','Maxiterations',Maxiter,'Tolerance',Tol);
A_Count_BW=normalise_rows(estA_Count_BW,Eps);
%%%%%%%%%%%%%%%%%%%%SKM
[estA1_Count_SKM, B_Count_BW_SKM] = hmmtrain(Training_Seq,A_Count,B_Count,'ALGORITHM','Viterbi','Maxiterations',Maxiter,'Tolerance',Tol,'Pseudotransitions',A_Count);
A_Count_SKM=normalise_rows(estA1_Count_SKM,Eps);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Method 3: Spectral Algorithm code by Carl Mattfeld
new_seq=split_seq(Training_Seq);%Represent observations as triples
%3A:SA AHK
[estB_SA_AHK, estA_SA_AHK] = learnAHK(new_seq,N,M1); %Learn AHK
A_SA_AHK=normalise_rows(estA_SA_AHK,Eps);
B_SA_AHK=normalise_rows(estB_SA_AHK,Eps);
%%%%%%%%%%%%%%%%%%%%BW
[estA1_SA_AHK_BW, B_SA_AHK_BW] = hmmtrain(Training_Seq,A_SA_AHK,B_SA_AHK,'ALGORITHM','BaumWelch','Maxiterations',Maxiter,'Tolerance',Tol);
A_SA_AHK_BW=normalise_rows(estA1_SA_AHK_BW,Eps);
B_SA_AHK_BW=normalise_rows(B_SA_AHK_BW,Eps);
%%%%%%%%%%%%%%%%%%%%SKM
%Repeat program when there is a Viterbi transition error
count=0; errors_SA_AHK_SKM=0;

while count==errors_SA_AHK_SKM
    try
        [estA1_SA_AHK_SKM, estB1_SA_AHK_SKM] = hmmtrain(Training_Seq,A_SA_AHK,B_SA_AHK,'ALGORITHM','Viterbi','Maxiterations',Maxiter,'Tolerance',Tol,'Pseudotransitions',A_SA_AHK);
        A_SA_AHK_SKM=normalise_rows(estA1_SA_AHK_SKM,Eps);
        B_SA_AHK_SKM=normalise_rows(estB1_SA_AHK_SKM,Eps);
    catch
        errors_SA_AHK_SKM=errors_SA_AHK_SKM+1;
    end
    count=count+1;
end
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%3B:SA HMM
[~, ~, ~, estB_SA_HMM, ~, estA_SA_HMM]...
    = learnHMM(new_seq,N,M1);%Learn HMM
%Obtain A and B using estimates from SA
absA_SA_HMM = abs(estA_SA_HMM);%Get magnitude values only
absB_SA_HMM = abs(estB_SA_HMM);%Get magnitude values only
A_SA_HMM=normalise_rows(absA_SA_HMM,Eps);%Check for zero or negative values and scale
B_SA_HMM=normalise_rows(absB_SA_HMM,Eps);%Check for zero or negative values and scale
%%%%%%%%%%%%%%%%%%%%BW
[estA1_SA_HMM_BW, estB1_SA_HMM_BW] = hmmtrain(Training_Seq,A_SA_HMM,B_SA_HMM,'ALGORITHM','BaumWelch','Maxiterations',Maxiter,'Tolerance',Tol);
A_SA_HMM_BW=normalise_rows(estA1_SA_HMM_BW,Eps);
B_SA_HMM_BW=normalise_rows(estB1_SA_HMM_BW,Eps);
%%%%%%%%%%%%%%%%%%%%SKM
count=0; errors_SA_HMM_SKM=0;
%Repeat program when there is a Viterbi transition error
while count==errors_SA_HMM_SKM
    try
        [estA1_SA_HMM_SKM, estB1_SA_HMM_SKM] = hmmtrain(Training_Seq,A_SA_HMM,B_SA_HMM,'ALGORITHM','Viterbi','Maxiterations',Maxiter,'Tolerance',Tol,'Pseudotransitions',A_SA_HMM);
        A_SA_HMM_SKM=normalise_rows(estA1_SA_HMM_SKM,Eps);
        B_SA_HMM_SKM=normalise_rows(estB1_SA_HMM_SKM,Eps);
    catch     
        errors_SA_HMM_SKM=errors_SA_HMM_SKM+1;
    end
    count=count+1;
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Method 4: Differential Evolution-SL (Hybrid)
[EstA_DE,EstB_DE]=differential_evolution(A_Rand,B_Rand,Training_Seq,Eps);%Obtain A and B using SL
A_DE=normalise_rows(EstA_DE,Eps);
B_DE=normalise_rows(EstB_DE,Eps);

%%%%%%%%%%%%%%%%%%%%BW
[estA1_DE_BW, B_DE_BW] = hmmtrain(Training_Seq,A_DE,B_DE,'ALGORITHM','BaumWelch','Maxiterations',Maxiter,'Tolerance',Tol);
A_DE_BW=normalise_rows(estA1_DE_BW,Eps);
%%%%%%%%%%%%%%%%%%%%SKM
[estA1_DE_SKM, B_DE_SKM] = hmmtrain(Training_Seq,A_DE,B_DE,'ALGORITHM','Viterbi','Maxiterations',Maxiter,'Tolerance',Tol,'Pseudotransitions',A_DE);
A_DE_SKM=normalise_rows(estA1_DE_SKM,Eps);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Method 5: Classical K-Means
%5A: Default-Entire Dataset
EstB_KmeansD=k_means_default(M1,N,Training_Seq');
B_KmeansD=normalise_rows(EstB_KmeansD,Eps);

%%%%%%%%%%%%%%%%%%%%BW
[estA1_KmeansD_BW, B_KmeansD_BW] = hmmtrain(Training_Seq,A_Rand,B_KmeansD,'ALGORITHM','BaumWelch','Maxiterations',Maxiter,'Tolerance',Tol);
A_KmeansD_BW=normalise_rows(estA1_KmeansD_BW,Eps);
%%%%%%%%%%%%%%%%%%%%SKM
[estA1_KmeansD_SKM, B_KmeansD_SKM] = hmmtrain(Training_Seq,A_Rand,B_KmeansD,'ALGORITHM','Viterbi','Maxiterations',Maxiter,'Tolerance',Tol,'Pseudotransitions',A_Rand);
A_KmeansD_SKM=normalise_rows(estA1_KmeansD_SKM,Eps);

%5B: Dataset Phases
EstB_KmeansP=k_means_phases(M1,N,Training_Data);
B_KmeansP=normalise_rows(EstB_KmeansP,Eps);
%%%%%%%%%%%%%%%%%%%%BW
[estA1_KmeansP_BW, B_KmeansP_BW] = hmmtrain(Training_Seq,A_Rand,B_KmeansP,'ALGORITHM','BaumWelch','Maxiterations',Maxiter,'Tolerance',Tol);
A_KmeansP_BW=normalise_rows(estA1_KmeansP_BW,Eps);
B_KmeansP_BW=normalise_rows(B_KmeansP_BW,Eps);

%%%%%%%%%%%%%%%%%%%%SKM
[estA1_KmeansP_SKM, B_KmeansP_SKM] = hmmtrain(Training_Seq,A_Rand,B_KmeansP,'ALGORITHM','Viterbi','Maxiterations',Maxiter,'Tolerance',Tol,'Pseudotransitions',A_Rand);
A_KmeansP_SKM=normalise_rows(estA1_KmeansP_SKM,Eps);
B_KmeansP_SKM=normalise_rows(B_KmeansP_SKM,Eps);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Part 5: Evaluation of Accuracy using A, B and Alpha for prediction(Method 1)                                             
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fprintf('Step 3 of 4:Evaluation using A, B and Alpha for prediction of NS and NO(Method 1)\n');
Results_Method1=["Sample Size" "Pred Appr" "SL" "SL_BW" "SL_SKM" "SA_AHK" "SA_AHK_BW" "SA_AHK_SKM" "SA_HMM" "SA_HMM_BW" "SA_HMM_SKM" "DE" "DE_BW" "DE_SKM"  "KmeansD" "KmeansD_BW" "KmeansD_SKM" "KmeansP" "KmeansP_BW" "KmeansP_SKM" "Unif_BW" "Unif_SKM" "Rand_BW" "Rand_SKM" "Count_BW" "Count_SKM"];
Observ_Symbols=unique(Inside1_Symbols(:,11));
Method =1;
for sSize=Start:Step:Stop   
    Acc1=evaluate(Method,A_Rand,Observ_Symbols,sSize,Evaluation_Data,A_SL, B_SL,A_SL_BW,B_SL_BW,A_SL_SKM,B_SL_SKM, A_Unif_BW,B_Unif_BW,A_Unif_SKM,B_Unif_SKM,A_Rand_BW,B_Rand_BW,A_Rand_SKM,B_Rand_SKM,A_Count_BW,B_Count_BW,A_Count_SKM,B_Count_BW_SKM,A_SA_AHK, B_SA_AHK,A_SA_AHK_BW,B_SA_AHK_BW,A_SA_AHK_SKM,B_SA_AHK_SKM,A_SA_HMM, B_SA_HMM,A_SA_HMM_BW,B_SA_HMM_BW,A_SA_HMM_SKM,B_SA_HMM_SKM,A_DE,B_DE,A_DE_BW,B_DE_BW,A_DE_SKM,B_DE_SKM,B_KmeansD,A_KmeansD_BW,B_KmeansD_BW,A_KmeansD_SKM,B_KmeansD_SKM,B_KmeansP,A_KmeansP_BW,B_KmeansP_BW,A_KmeansP_SKM,B_KmeansP_SKM);
    Results_Method1=[Results_Method1;Acc1];
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Part 6: Evaluation of Accuracy using A and Alpha for NS, and B and Alpha for NO prediction(Method 2)         %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fprintf('Step 4 of 4:Evaluation using A and Alpha for NS, and B and Alpha for NO prediction(Method 2)\n');
Results_Method2=["Sample Size" "Pred Appr" "SL" "SL_BW" "SL_SKM" "SA_AHK" "SA_AHK_BW" "SA_AHK_SKM" "SA_HMM" "SA_HMM_BW" "SA_HMM_SKM" "DE" "DE_BW" "DE_SKM"  "KmeansD" "KmeansD_BW" "KmeansD_SKM" "KmeansP" "KmeansP_BW" "KmeansP_SKM" "Unif_BW" "Unif_SKM" "Rand_BW" "Rand_SKM" "Count_BW" "Count_SKM"];
Method=2;
for sSize=Start:Step:Stop    
    Acc2=evaluate(Method,A_Rand,Observ_Symbols,sSize,Evaluation_Data,A_SL, B_SL,A_SL_BW,B_SL_BW,A_SL_SKM,B_SL_SKM, A_Unif_BW,B_Unif_BW,A_Unif_SKM,B_Unif_SKM,A_Rand_BW,B_Rand_BW,A_Rand_SKM,B_Rand_SKM,A_Count_BW,B_Count_BW,A_Count_SKM,B_Count_BW_SKM,A_SA_AHK, B_SA_AHK,A_SA_AHK_BW,B_SA_AHK_BW,A_SA_AHK_SKM,B_SA_AHK_SKM,A_SA_HMM, B_SA_HMM,A_SA_HMM_BW,B_SA_HMM_BW,A_SA_HMM_SKM,B_SA_HMM_SKM,A_DE,B_DE,A_DE_BW,B_DE_BW,A_DE_SKM,B_DE_SKM,B_KmeansD,A_KmeansD_BW,B_KmeansD_BW,A_KmeansD_SKM,B_KmeansD_SKM,B_KmeansP,A_KmeansP_BW,B_KmeansP_BW,A_KmeansP_SKM,B_KmeansP_SKM);
    Results_Method2=[Results_Method2;Acc2];
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Part 7:   Plotting results
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
plotting_results(Results_Method1,Results_Method2,Start,Step,Stop);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                 End of Main Program                                       
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%